/* Copyright (C) 2009 Sakse Dalum
 * 
 * This file is part of Poddie Server.
 * 
 * Poddie Server is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Poddie Server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * A copy of the GNU Affero General Public License can be found in the
 * file LICENSE. 
 */

/*
 *                         player.cpp  -  description
 *                         --------------------------------
 *   begin                : Tue Aug 13 2009
 *   copyright            : (C) 2009 by Sakse Dalum
 *   maintainer(s)        : Sakse "Don S" Dalum <don.s@yodanism.org>
 *   description          : Player object to control player variables,
 *                          components, etc.
 *
 */

#include "player.h"

/***********************************************************************
*    *    *    *    *                              *    *    *    *    *
                            CON- & DESTRUCTORS
*    *    *    *    *                              *    *    *    *    *
***********************************************************************/

Player::Player(float x, float y, float angle, World *world, int cid)
: GameObject(x, y, angle, world), cid(cid) {
	alive = false;
	
	acceleration = 0.2;
	turnSpeed = 0;
	doesTurn = false;
	turnRight = false;
	doesMove = false;
	
	timer_last_update = CL_System::get_time();
	timer_respawn = CL_System::get_time() / 1000.0f + 5.0f;
	do_respawn = true;
	
	radius = 500;
	
	health = 100;
	
	for (int i = 0; i < 16; i++) slot[i] = false;
}

Player::Player() {}

Player::~Player() {
	
}

/***********************************************************************
*    *    *    *    *                              *    *    *    *    *
                            LOOPING FUNCTIONS
*    *    *    *    *                              *    *    *    *    *
***********************************************************************/

void Player::update() {
	
	// Do the basic physics & movement calculations.
	angle += turnSpeed;
	x += hspeed;
	y += vspeed;
	apply_friction(1.005);
	
	// Update our components.
	std::list<Component*>::iterator it; 
	for (it = components.begin(); it != components.end(); it++) {
		Component *cur = *it;
		cur->update();
	}
	
	// Disconnect the player after 20 seconds of missing update packets.
	if (timer_last_update+20000 <= CL_System::get_time()) {
		world->disconnect_player(cid);
	}
	
	/// //////////////////////////////////////////////////////////// ///
	///                       TEMPORARY CODE                         ///
	/// //////////////////////////////////////////////////////////// ///
	
	//! TODO: Rewrite a proper border-handling function.
	if (x < world->border_left) { x = world->border_left; hspeed = 0; }
	if (x > world->border_right) { x = world->border_right; hspeed = 0; }
	if (y < world->border_top) { y = world->border_top; vspeed = 0; }
	if (y > world->border_bottom) { y = world->border_bottom; vspeed = 0; }
	
	
	//! TODO: Rewrite a proper respawn function.
	if (health <= 0) {
		health = 100;
		timer_respawn = CL_System::get_time() / 1000.0f + 5.0f;
		do_respawn = true;
	}
	
	if (timer_respawn < CL_System::get_time() / 1000.0f && do_respawn) {
		respawn();
		do_respawn = false;
	}
	else if (do_respawn) {
		x = -100000;
		y = -100000;
	}
}

/***********************************************************************
*    *    *    *    *                              *    *    *    *    *
                               FUNCTIONS
*    *    *    *    *                              *    *    *    *    *
***********************************************************************/


////////////////////////////////////////////////////////////////////////
//                     MOVEMENT, POSITION & DIRECTION

void Player::set_position(float new_x, float new_y) {
	x = new_x;
	y = new_y;
}

void Player::turn_right() {
	turnRight = true;
}

void Player::turn_left() {
	turnRight = false;
}

void Player::respawn() {
	x = (float)(rand() % (2048*3*15));
	y = (float)(rand() % (2048*3*15));
	hspeed = 0;
	vspeed = 0;
	turnSpeed = 0;
}                   

//                           SIMPLE PHYSICS
void Player::add_vector(float angle, float acceleration) {
	vspeed += sin(angle*DEG_TO_RAD)*acceleration;
	hspeed += cos(angle*DEG_TO_RAD)*acceleration;
}

void Player::add_turn(float speed) {
	turnSpeed += speed;
}

void Player::apply_friction(float friction) {
	vspeed /= friction;
	hspeed /= friction;
	turnSpeed /= friction;
}

////////////////////////////////////////////////////////////////////////
//                             COMPONENTS

//                                ADD
void Player::add_component(list<Component*>::iterator pos, Component *comp) {
	components.insert(pos, comp);
	world->network_component_add_send(cid, comp->type, comp->part, comp->number, comp->slot);
	slot[comp->slot] = true;
	component[comp->slot] = --components.end();
}

//                               REMOVE
void Player::remove_component(list<Component*>::iterator pos, int slot) {
	components.erase(pos);
	world->network_component_remove_send(cid, slot);
	this->slot[slot] = false;
}

////////////////////////////////////////////////////////////////////////
//                             CONTROLLER

void Player::set_controller(std::string new_controller) {
	controller = new_controller;
}

std::string Player::get_controller() {
	return controller;
}
