/* Copyright (C) 2009 Sakse Dalum
 * 
 * This file is part of Poddie Server.
 * 
 * Poddie Server is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Poddie Server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * A copy of the GNU Affero General Public License can be found in the
 * file LICENSE. 
 */

/*
 *                         world.cpp  -  description
 *                         --------------------------------
 *   begin                : Tue Aug 13 2009
 *   copyright            : (C) 2009 by Sakse Dalum
 *   maintainer(s)        : Sakse "Don S" Dalum <don.s@yodanism.org>
 *   description          : Controls all major functions of the game,
 *                          and takes care of updating objects etc.
 *
 */

#include "world.h"

/***********************************************************************
*    *    *    *    *                              *    *    *    *    *
                            CON- & DESTRUCTORS
*    *    *    *    *                              *    *    *    *    *
***********************************************************************/

//World constructor
World::World(string port) : port(port) {
	players_connected = 0;
	
	//Map borders
	border_left = 500;
	border_right = (2048*3*15)-500;
	border_top = 500;
	border_bottom = (2048*3*15)-500;
	
	//Update counters.
	timer_update_player = 0.0f;
	timer_update_missile = 0.0f;
	
	//If true, the game will end.
	quit = false;
}
 
//World destructor
World::~World() {	
}

/***********************************************************************
*    *    *    *    *                              *    *    *    *    *
                          THE ACTUAL GAME LOOP
*    *    *    *    *                              *    *    *    *    *
***********************************************************************/

void World::run() {
	////////////////////////////////////////////////////////////////
	//                       LOADING & SETUP
	
	cout << "Creating netsession... \t";

	//Create the netsession.
	netsession = new CL_NetSession("Poddie");
	
	cout << "Success." << endl;
	cout << "Now listening on port: " << port << endl;

	netsession->start_listen(port);
	
	cout << "Start listening for packets... \t";
	
	//Listen for packets on the following channels:
	slots.connect(netsession->sig_netpacket_receive("component_add"), this, &World::network_component_add_receive);
	slots.connect(netsession->sig_netpacket_receive("component_remove"), this, &World::network_component_remove_receive);
	slots.connect(netsession->sig_netpacket_receive("component_activate"), this, &World::network_component_activate_receive);
	slots.connect(netsession->sig_netpacket_receive("components_request"), this, &World::network_components_request_receive);
	slots.connect(netsession->sig_netpacket_receive("delay_return"), this, &World::network_delay_return_receive);
	slots.connect(netsession->sig_netpacket_receive("disconnect"), this, &World::network_disconnect_receive);
	
	//Connect and disconnect signals.
	slots.connect(netsession->sig_computer_connected(), this, &World::network_connect);
	slots.connect(netsession->sig_computer_disconnected(), this, &World::network_disconnect);

	cout << "Success." << endl;
	
	cout << "Server is running and open for connections..." << endl;
	
	////////////////////////////////////////////////////////////////////
	//                         GAME LOOP
	
	while (!quit) {
		
		//--------------------------------------------------------------
		//!               UPDATE GAME OBJECTS
		//--------------------------------------------------------------
		
		//!                    MISSILES
		try {
			for (int i = 0; i < missile.size(); i++) {
				if (missile.at(i)->alive) {
					
					// Check for collisions with player objects.
					for (int cid = 0; cid < player.size(); cid++) {
						if (player.at(cid)->alive) {
							
							float x = player.at(cid)->get_x();
							float y = player.at(cid)->get_y();
							float radius = player.at(cid)->radius+missile.at(i)->radius;
							
							if (missile.at(i)->collision_circle(x, y, radius)
								&& missile.at(i)->creator != player.at(cid)) {
								missile.at(i)->alive = false;
								
								// Send packet telling that a collision occurred.
								network_collision_send(cid, PLAYER);
								network_collision_send(i, MISSILE);
								
								// Tell the client that the missile is dead.
								network_object_status_send(i, MISSILE);
								
								// Do whatever we have to do with the player.
								player.at(cid)->health -= 10;
							}
						}
					}
					
					// Update the missile.
					missile.at(i)->update();
					
					if (CL_System::get_time() / 1000.0f > timer_update_missile) {
						// Send updates about the missile to all players.
						network_object_status_send(i, MISSILE);
					}
				}
			}
		}
		catch (Exception &e) {
			cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
		}
		
		//!                     PLAYERS
		try {
			for (int cid = 0; cid < player.size(); cid++) {
				if (player.at(cid)->alive) {
					
					// Update the player.
					player.at(cid)->update();
					
					if (CL_System::get_time() / 1000.0f > timer_update_player) {
						//Send updates about the player to all players.
						network_object_status_send(cid, PLAYER);
					}
				}
			}
		}
		catch (Exception &e) {
			cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
		}
		
		//--------------------------------------------------------------
		//!                     TIMERS
		//--------------------------------------------------------------
		
		if (CL_System::get_time() / 1000.0f > timer_update_player)
			timer_update_player = CL_System::get_time() / 1000.0f + 0.02f;
			// Players will be updated every 0.02 second(s) (20ms). ^

		if (CL_System::get_time() / 1000.0f > timer_update_missile)
			timer_update_missile = CL_System::get_time() / 1000.0f + 1.0f;
			// Players will be updated every 1.0 second(s) (1000ms). ^
		
		//Keep the game alive and set the framerate to 100 (update rate = 10ms).
		CL_System::keep_alive(10);
	}
}


/***********************************************************************
*    *    *    *    *                              *    *    *    *    *
                               NETWORKING
*    *    *    *    *                              *    *    *    *    *
***********************************************************************/

//!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
//                          IN-GAME RELATED
//!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

////////////////////////////////////////////////////////////////////////
//                              SENDING

//----------------------------------------------------------------------
//!                  MOVEMENT, POSITION & DIRECTION
//----------------------------------------------------------------------

//                           OBJECT STATUS
void World::network_object_status_send(int id, int object) {	
	packet_object_status = new CL_NetPacket();
	
	if (object == MISSILE) {
		packet_object_status->output.write_bool8(missile.at(id)->alive);
		packet_object_status->output.write_int32(id);
		packet_object_status->output.write_int32(object);
		packet_object_status->output.write_float32(missile.at(id)->get_hspeed());
		packet_object_status->output.write_float32(missile.at(id)->get_vspeed());
		packet_object_status->output.write_float32(missile.at(id)->get_angle());
		packet_object_status->output.write_float32(missile.at(id)->get_x());
		packet_object_status->output.write_float32(missile.at(id)->get_y());
	}
	
	if (object == PLAYER) {
		packet_object_status->output.write_bool8(player.at(id)->alive);
		packet_object_status->output.write_int32(id);
		packet_object_status->output.write_int32(object);
		packet_object_status->output.write_float32(player.at(id)->get_hspeed());
		packet_object_status->output.write_float32(player.at(id)->get_vspeed());
		packet_object_status->output.write_float32(player.at(id)->get_angle());
		packet_object_status->output.write_float32(player.at(id)->get_x());
		packet_object_status->output.write_float32(player.at(id)->get_y());
	}
	netsession->get_all().send("object_status", *packet_object_status);
	delete packet_object_status;	// discard sent message
}

//----------------------------------------------------------------------
//!                             MISSILE
//----------------------------------------------------------------------

//                              CREATE
void World::network_missile_create_send(bool alive, int cid, float x, float y, float angle, float hspeed, float vspeed, string sprite) {
	packet_missile_create = new CL_NetPacket();
	packet_missile_create->output.write_bool8(alive);
	packet_missile_create->output.write_int32(cid);
	packet_missile_create->output.write_float32(x);
	packet_missile_create->output.write_float32(y);
	packet_missile_create->output.write_float32(angle);
	packet_missile_create->output.write_float32(hspeed);
	packet_missile_create->output.write_float32(vspeed);
	packet_missile_create->output.write_string(sprite);
	netsession->get_all().send("missile_create", *packet_missile_create);
	delete packet_missile_create; // discard sent message
}

//----------------------------------------------------------------------
//!                            COLLISION
//----------------------------------------------------------------------

void World::network_collision_send(int id, int object) {	
	packet_collision = new CL_NetPacket();
	
	if (object == MISSILE) {
		packet_collision->output.write_int32(id);
		packet_collision->output.write_int32(object);
		packet_collision->output.write_float32(missile.at(id)->radius);
	}
	
	if (object == PLAYER) {
		packet_collision->output.write_int32(id);
		packet_collision->output.write_int32(object);
		packet_collision->output.write_float32(player.at(id)->radius);
	}
	netsession->get_all().send("collision", *packet_collision);
	delete packet_collision; // discard sent message
}

//----------------------------------------------------------------------
//!                            COMPONENT
//----------------------------------------------------------------------

//                                ADD
void World::network_component_add_send(int cid, string type, string part, string number, int slot) {
	packet_component_add = new CL_NetPacket();
	packet_component_add->output.write_int32(cid);
	packet_component_add->output.write_string(type);
	packet_component_add->output.write_string(part);
	packet_component_add->output.write_string(number);
	packet_component_add->output.write_int32(slot);
	netsession->get_all().send("component_add", *packet_component_add);
	delete packet_component_add;	// discard sent message
}

//                              REMOVE
void World::network_component_remove_send(int cid, int slot) {
	packet_component_remove = new CL_NetPacket();
	packet_component_remove->output.write_int32(cid);
	packet_component_remove->output.write_int32(slot);
	netsession->get_all().send("component_remove", *packet_component_remove);
	delete packet_component_remove;	// discard sent message
}

//                             ACTIVATE
void World::network_component_activate_send(int cid, string func, bool active, int slot) {	
	packet_component_activate = new CL_NetPacket();
	packet_component_activate->output.write_int32(cid);
	packet_component_activate->output.write_string(func);
	packet_component_activate->output.write_bool8(active);
	packet_component_activate->output.write_int32(slot);
	netsession->get_all().send("component_activate", *packet_component_activate);
	delete packet_component_activate;	// discard sent message
}

//                             ACTIVITY
void World::network_component_activity_send(int cid, float activity, int slot) {
	packet_component_activity = new CL_NetPacket();
	packet_component_activity->output.write_int32(cid);
	packet_component_activity->output.write_float32(activity);
	packet_component_activity->output.write_int32(slot);
	netsession->get_all().send("activity", *packet_component_activity);
	delete packet_component_activity;	// discard sent message
}

////////////////////////////////////////////////////////////////////////
//                             RECEIVING

//----------------------------------------------------------------------
//!                           COMPONENT(S)
//----------------------------------------------------------------------

//                                ADD
void World::network_component_add_receive(CL_NetPacket &packet, CL_NetComputer &computer) {
	try {
		int cid = packet.input.read_int32();
		string ip = computer.get_address().get_address();
		
		// Check if the Player's IP matches the IP the packet was sent from.
		if (player.at(cid)->get_controller() == ip && player.at(cid)->alive) {
			
			// Read the packet contents.
			string type = packet.input.read_string();
			string part = packet.input.read_string();
			string number = packet.input.read_string();
			int slot = packet.input.read_int32();
			
			if (slot < sizeof(player.at(cid)->slot) && slot >= 0) {
				if (!player.at(cid)->slot[slot]) {
					cout << "Player " << cid << " sent request to add component: \"";
					cout << type << "/" << part << "/" << number << endl;
					// Add the component.
					Component *comp;
					comp = new Component(slot, type, part, number, player.at(cid), this);
					player.at(cid)->add_component(player.at(cid)->components.end(), comp);
				}
			}
		}
	}
	catch (Exception &e) {
		cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
	}
}

//                              REMOVE
void World::network_component_remove_receive(CL_NetPacket &packet, CL_NetComputer &computer) {
	try {
		int cid = packet.input.read_int32();
		string ip = computer.get_address().get_address();
		
		// Check if the Player's IP matches the IP the packet was sent from.
		if (player.at(cid)->get_controller() == ip && player.at(cid)->alive) {
			
			// Read the packet.
			int slot = packet.input.read_int32();
			
			if (slot < sizeof(player.at(cid)->slot) && slot >= 0) {
				if (player.at(cid)->slot[slot]) {
					cout << "Player " << cid << " sent request to remove component at slot: " << slot << endl;
					
					// Remove the component.
					player.at(cid)->remove_component(player.at(cid)->component[slot], slot);
				}
			}
		}
	}
	catch (Exception &e) {
		cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
	}
}

//                             ACTIVATE
void World::network_component_activate_receive(CL_NetPacket &packet, CL_NetComputer &computer) {
	try {
		int cid = packet.input.read_int32();
		string ip = computer.get_address().get_address();
		
		// Check if the Player's IP matches the IP the packet was sent from.
		if (player.at(cid)->get_controller() == ip && player.at(cid)->alive) {
			
			// Read the packet.
			string func = packet.input.read_string();
			bool state = packet.input.read_bool8();
			int slot = packet.input.read_int32();
			
			// Determine which variable we need to change the state of.
			list<Component*>::iterator it; 
			for (it = player.at(cid)->components.begin(); it != player.at(cid)->components.end(); it++) {
				Component *cur = *it;
				if (cur->slot == slot) {
					if (func == "active") cur->active = state;
					if (func == "increase") cur->increase = state;
					if (func == "decrease") cur->decrease = state;
					
					// Send the updated state to all players.
					network_component_activate_send(cid, func, state, slot);
				}
			}
		}
	}
	catch (Exception &e) {
		cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
	}
}

//                              REQUEST
void World::network_components_request_receive(CL_NetPacket &packet, CL_NetComputer &computer) {
	try {
		int cid = packet.input.read_int32();
		string ip = computer.get_address().get_address();
		
		// Check if the Player's IP matches the IP the packet was sent from.
		if (player.at(cid)->get_controller() == ip && player.at(cid)->alive) {
			
			for (int i = 0; i < player.size(); i++) {
				
				// Only send components of alive players.
				if (player.at(i)->alive) {
					
					list<Component*>::iterator it; 
					for (it = player.at(i)->components.begin(); it != player.at(i)->components.end(); it++) {
						
						Component *cur = *it;
						packet_component_add = new CL_NetPacket();
						packet_component_add->output.write_int32(player.at(i)->cid);
						packet_component_add->output.write_string(cur->type);
						packet_component_add->output.write_string(cur->part);
						packet_component_add->output.write_string(cur->number);
						packet_component_add->output.write_int32(cur->slot);
						computer.send("component_add", *packet_component_add);
						delete packet_component_add; // discard sent message
						
						cout << "Sent packet: " << player.at(i)->cid << ", ";
						cout << cur->type << ", " << cur->part << ", ";
						cout << cur->number << ", " << cur->slot << endl;
					}
				}
			}
		}
	}
	catch (Exception &e) {
		cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
	}
}

//----------------------------------------------------------------------
//!                              DELAY
//----------------------------------------------------------------------

void World::network_delay_return_receive(CL_NetPacket &packet, CL_NetComputer &computer) {
	try {
		int cid = packet.input.read_int32();
		string ip = computer.get_address().get_address();
		
		// Check if the Player's IP matches the IP the packet was sent from.
		if (player.at(cid)->get_controller() == ip && player.at(cid)->alive) {
			
			packet_delay = new CL_NetPacket();
			packet_delay->output.write_int32(cid);
			computer.send("delay", *packet_delay);
			delete packet_delay; // discard sent message
			
			player.at(cid)->timer_last_update = CL_System::get_time();
		}
	}
	catch (Exception &e) {
		cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
	}
}

//!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
//                         CONNECTION RELATED
//!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

////////////////////////////////////////////////////////////////////////
//                            CONNECTIONS

//----------------------------------------------------------------------
//!                             CONNECT
//----------------------------------------------------------------------

void World::network_connect(CL_NetComputer &computer) {
	try {
		cout << computer.get_address().get_address() << " connected." << endl;
		players_connected++;
		
		// Tell the newly connected client which players exist.
		for (int i = 0; i < player.size(); i++) {
			CL_NetPacket *msg;
			msg = new CL_NetPacket();
			msg->output.write_int32(i);
			msg->output.write_bool8(false);
			if (player.at(i)->alive) msg->output.write_bool8(true);
			else msg->output.write_bool8(false);
			computer.send("alive", *msg);
			delete msg;	// discard sent message
		}
		
		// Tell the newly connected client which objects exist.
		for (int i = 0; i < missile.size(); i++) {
			packet_missile_create = new CL_NetPacket();
			packet_missile_create->output.write_bool8(missile.at(i)->alive);
			packet_missile_create->output.write_int32(missile.at(i)->creator->cid);
			packet_missile_create->output.write_float32(missile.at(i)->x);
			packet_missile_create->output.write_float32(missile.at(i)->y);
			packet_missile_create->output.write_float32(missile.at(i)->angle);
			packet_missile_create->output.write_float32(missile.at(i)->hspeed);
			packet_missile_create->output.write_float32(missile.at(i)->vspeed);
			packet_missile_create->output.write_string(missile.at(i)->sprite);
			computer.send("missile_create", *packet_missile_create);
			delete packet_missile_create;	// discard sent message
		}
		
		// Create a new player for our newly connected client.
		int cid = player.size();
		Player *new_player;
		new_player = new Player((2048*3*15)/2, (2048*3*15)/2, 0, this, cid);
		player.push_back(new_player);
		
		cout << "Total amount of players since last startup: " << player.size() << endl;
		cout << "Players connected: " << players_connected << endl;
		
		player.at(cid)->set_controller(computer.get_address().get_address());
		player.at(cid)->alive = true;
		
		//Tell all clients that the new player is alive.
		CL_NetPacket *msg;
		msg = new CL_NetPacket();
		msg->output.write_int32(cid);
		/* We tell all clients, that they control this player. It's now
		 * up to the client to decide whether that is true or not. This
		 * won't affect gameplay, unless you use a modified client, that
		 * handles player control differently. */
		msg->output.write_bool8(true);
		msg->output.write_bool8(true);
		netsession->get_all().send("alive", *msg);
		delete msg;	// discard sent message
	}
	catch (Exception &e) {
		cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
	}
}

//----------------------------------------------------------------------
//!                            DISCONNECT
//----------------------------------------------------------------------

//                               SEND
void World::network_disconnect_player_send(int cid) {
	packet_disconnect_player = new CL_NetPacket();
	packet_disconnect_player->output.write_int32(cid);
	netsession->get_all().send("dead", *packet_disconnect_player);
	delete packet_disconnect_player; // discard sent message
}

//                              RECEIVE
void World::network_disconnect_receive(CL_NetPacket &packet, CL_NetComputer &computer) {
	try {
		cout << "Received disconnect signal." << endl;
	
		int cid = packet.input.read_int32();
		
		cout << "Read player id: " << cid << endl;
		
		string ip = computer.get_address().get_address();
		
		cout << "Read player IP: " << ip << endl;
		
		// Check if the Player's IP matches the IP the packet was sent from.
		if (player.at(cid)->get_controller() == ip && player.at(cid)->alive) {
			
			// Disconnect the player.
			disconnect_player(cid);
		}
		else cout << "Received invalid request to disconnect Player " << cid <<  " from "<< ip << endl;
	}
	catch (Exception &e) {
		cerr << "Error: " << e.message() << " (" << e.what() << ")" << endl;
	}
}

//                     FUNCTION TO DISCONNECT A PLAYER
void World::disconnect_player(int cid) {
	
	// Tell the clients that the player has been disconnected.
	network_disconnect_player_send(cid);
	
	player.at(cid)->alive = false;
	
	// Player have been disconnected - we don't want to allow him to mess
	// with the server, so we remove his IP.
	player.at(cid)->set_controller(""); 
	
	cout << "Player " << cid << " left the game." << endl;
	
	players_connected--;
	
	cout << "Players connected: " << players_connected << endl;
}

//                       ACTUAL DISCONNECTION
void World::network_disconnect(CL_NetComputer &computer) {
	cout << computer.get_address().get_address() << " disconnected." << endl;
}
